static char rcsid[] = "@(#)$Id: service_list.c,v 2.16 2015/12/01 20:15:46 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.16 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                  (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "elm_defs.h"
#include "ss_imp.h"
#include "connection_imp.h"
#ifdef USE_DLOPEN
#include "shared_imp.h"
#endif
#include "s_me.h"
#include "s_elm.h"
#include "rc_imp.h"

DEBUG_VAR(Debug,__FILE__,"net");

#ifdef REMOTE_MBX

/* Seems that h_errno is macro on AIX */
#ifndef h_errno
extern int h_errno;
#endif

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif

static const unsigned char *cs2us P_((const char *str));
static const unsigned char *cs2us(str) 
     const char *str;
{
    return (const unsigned char *)str;
}

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

static char *us2s P_((unsigned char *str));
static char *us2s(str) 
     unsigned char *str;
{
    return (char *)str;
}

/* Hard coded */
static struct service_type {
    const  char * name;
    int           flags;
    PORTS         defport;
    const char  * have_init_tls; /* message text for 
				    remote_account_init_tls */
} SERVICE_TYPES[] = {
    { "*",       
      STFLAG_browser|STFLAG_mbox|
      STFLAG_is_imap|STFLAG_is_pop |
      STFLAG_smtp_like|
      STFLAG_is_submission|STFLAG_is_smtp,
      PORT_end, NULL},
    { "imap",    STFLAG_browser|STFLAG_mbox|STFLAG_is_imap,    PORT_imap4, 
      NULL
    },
    { "pop",     STFLAG_mbox|STFLAG_is_pop,                    PORT_pop3,
      NULL
    },

    { "submission", STFLAG_smtp_like|STFLAG_is_submission,     PORT_submission, 
      NULL
    },
    { "smtp",    STFLAG_smtp_like|STFLAG_is_smtp,              PORT_smtp,
      NULL
    },
    { "smtps",    STFLAG_smtp_like|STFLAG_require_port, 
      PORT_end  /* no port given */,
      "SMTPS"
    },


    { "imaps",   STFLAG_browser|STFLAG_mbox|STFLAG_is_imap,    PORT_imaps,
      "IMAPs"
    },
    { "pops",    STFLAG_mbox|STFLAG_is_pop,                    PORT_pop3s,
      "POPs"
    },

      
    { NULL,      0,                                            PORT_end, 
      NULL
    },
};



static struct service_entry * malloc_service_entry P_((const char *name,
						       const struct service_type *st,
						       int flags));

static struct service_entry * malloc_service_entry(name,st,flags)
     const char *name;
     const struct service_type *st;
     int flags;
{
    struct service_entry * entry = safe_malloc(sizeof (*entry));

    bzero((void *)entry,sizeof (*entry));

    entry->refcount       = 1;
    
    entry->flags          = flags;           
    entry->official_name  = safe_strdup(name);
    entry->aliases_list   = 0;
    entry->aliases_count  = 0;
    entry->addr_list      = NULL;
    entry->addr_count     = 0;

    entry->addr_name_list  = NULL;
    entry->addr_name_count = 0;

    entry->port_name_list  = NULL;
    entry->port_name_count = 0;

    entry->service        = st;
    /* ??? */
    entry->port_list      = NULL;
    entry->port_count     = 0;

    entry->option_list    = NULL;
    entry->option_count   = 0;

    entry->req_tls_peername_list    = NULL;
    entry->req_tls_peername_count   = 0;

    entry->magic = SERVICE_ENTRY_magic;
    return entry;
}

int  STFLAG_service_type(se)
     struct service_entry *se;
{
    if (SERVICE_ENTRY_magic != se->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "STFLAG_service_type",
	      "Bad magic (service_entry)",0);
    
    return se->service->flags & STFLAG_type_mask;
}

const char * have_init_tls(se)
     struct service_entry *se;
{
    if (SERVICE_ENTRY_magic != se->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "STFLAG_service_type",
	      "Bad magic (service_entry)",0);

    return se->service->have_init_tls;
}

void add_addr_name_to_entry(entry,name,malloc_it)
     struct service_entry *entry;
     char *name;
     int malloc_it;
{
    if (SERVICE_ENTRY_magic != entry->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "add_addr_name_to_entry",
		      "Bad magic (service_entry)",0);

    entry->addr_name_list = 
	safe_realloc(entry->addr_name_list,
		     (entry->addr_name_count+1) * 
		     sizeof (entry->addr_name_list[0]));
    entry->addr_name_list[entry->addr_name_count++]
	= malloc_it ? safe_strdup(name) : name;
}

void add_port_name_to_entry(entry,name,malloc_it)
     struct service_entry *entry;
     char *name;
     int malloc_it;
{
    if (SERVICE_ENTRY_magic != entry->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "add_port_name_to_entry",
		      "Bad magic (service_entry)",0);

    entry->port_name_list = 
	safe_realloc(entry->port_name_list,
		     (entry->port_name_count+1) * 
		     sizeof (entry->port_name_list[0]));
    entry->port_name_list[entry->port_name_count++]
	= malloc_it ? safe_strdup(name) : name;
}

void add_alias_to_entry(entry,name)
     struct service_entry *entry;
     const char *name;    
{
    if (SERVICE_ENTRY_magic != entry->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "add_alias_to_entry",
	      "Bad magic (service_entry)",0);

    entry->aliases_list = 
	safe_realloc(entry->aliases_list,
		     (entry->aliases_count+1) * 
		     sizeof (entry->aliases_list[0]));
    entry->aliases_list[entry->aliases_count++]
	= safe_strdup(name);
}

void add_addr_to_entry(entry,result)
     struct service_entry *entry;
     SOCKADDR *result;
{
    if (SERVICE_ENTRY_magic != entry->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "add_addr_to_entry",
	      "Bad magic (service_entry)",0);

    entry->addr_list = 
	safe_realloc(entry->addr_list,
		     (entry->addr_count+1) * 
		     sizeof (entry->addr_list[0]));
    bzero((void *)&(entry->addr_list[entry->addr_count]),
	  sizeof (entry->addr_list[entry->addr_count]));

    entry->addr_list[entry->addr_count++] = *result;
}

void add_port_to_entry(entry,port)
     struct service_entry *entry;
     int port;    
{
    if (SERVICE_ENTRY_magic != entry->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "add_port_to_entry",
	      "Bad magic (service_entry)",0);

    entry->port_list = 
	safe_realloc(entry->port_list,
		     (entry->port_count+2) * 
		     sizeof (entry->port_list[0]));

    entry->port_list[entry->port_count++] = port;
    entry->port_list[entry->port_count]   = PORT_end;
}

int add_option_type_to_entry(entry,Y,prefix)
     struct service_entry *entry;
     struct SE_option_type * Y;
     const char * prefix;
{
    if (SERVICE_ENTRY_magic != entry->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "add_option_type_to_entry",
	      "Bad magic (service_entry)",0);

    entry->option_list =
	safe_realloc(entry->option_list,
		     (entry->option_count+1) *
		     sizeof (entry->option_list[0]));
    bzero((void *) &(entry->option_list[entry->option_count]),
	   sizeof (entry->option_list[0]));

    entry->option_list[entry->option_count].type   = Y;
    entry->option_list[entry->option_count].prefix = safe_strdup(prefix);
    entry->option_list[entry->option_count].value = NULL;

    Y->zero_options(& (entry->option_list[entry->option_count]));
    return entry->option_count++;
}

void add_req_peername_to_entry(entry,req_peername)
     struct service_entry *entry;
     const struct string * req_peername;
{
    int i;

    if (SERVICE_ENTRY_magic != entry->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "add_req_peername_to_entry",
	      "Bad magic (service_entry)",0);

    for (i = 0; i < entry->req_tls_peername_count; i++) {
	if (0 == string_cmp(entry->req_tls_peername_list[i],
			    req_peername,-99 /* Unknown value*/))
	    break;

    }

    if (i >= entry->req_tls_peername_count) {
	entry->req_tls_peername_list
	    = safe_realloc(entry->req_tls_peername_list,
			   sizeof (entry->req_tls_peername_list[0]) *
			   (entry->req_tls_peername_count +1));

	entry->req_tls_peername_list[entry->req_tls_peername_count++]
	    = dup_string(req_peername);
    }
}



    

static void merge1_entries P_((struct service_entry *target,
			       struct service_entry *value));
static void merge1_entries(target,value)
     struct service_entry *target;
     struct service_entry *value;
{
    if (SERVICE_ENTRY_magic != target->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "merge1_entries",
	      "Bad magic (service_entry) (target)",0);
    if (SERVICE_ENTRY_magic != value->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "merge1_entries",
	      "Bad magic (service_entry) (value)",0);

    if (target->service == &(SERVICE_TYPES[0])) {
	target->service = value->service;
	if (target->service != &(SERVICE_TYPES[0])) {
	    DPRINT(Debug,10,(&Debug,  "... setting temporary entry for %s to be type %s\n",
			     target->official_name,
			     target->service->name));
	}
    } else if (target->service != value->service &&
	       value->service != &(SERVICE_TYPES[0])) {
	DPRINT(Debug,1,(&Debug,  
			 "... conflicting service entry types. Type %s used, %s ignored\n",
			target->service->name,
			value->service->name
			));
	return;
    }

    
    if (0 != istrcmp(target->official_name,value->official_name)) {
	int j;

	for (j = 0; j < target->aliases_count; j++) {
	    if (0 == istrcmp(value->official_name,target->aliases_list[j]))
		break;
	}
	if (j >= target->aliases_count) 
	    add_alias_to_entry(target,value->official_name);
	
    }

    /* Add missing aliases */
    if ((value->flags & SE_given_alias) &&
	value->aliases_count > 0) {
	int i;

	target->flags |= SE_given_alias;

	for (i = 0; i < value->aliases_count; i++) {
	    int j;
	    if (0 == istrcmp(target->official_name,value->aliases_list[i]))
		continue;

	    for (j = 0; j < target->aliases_count; j++) {
		if (0 == istrcmp(value->aliases_list[i],
				 target->aliases_list[j]))
		    break;
	    }
	    if (j >= target->aliases_count) 
		add_alias_to_entry(target,value->aliases_list[i]);
	}
    }
	
    if ((value->flags & SE_given_name_addr) &&
	value->addr_name_count > 0) {
	int i;
	target->flags |= SE_given_name_addr;

	for (i = 0; i < value->addr_name_count; i++) {
	    int j;

	    for (j = 0; j < target->addr_name_count; j++) {
		if (0 == istrcmp(value->addr_name_list[i],
				 target->addr_name_list[j]))
		    break;		
	    }
	    if (j >= target->addr_name_count)
		add_addr_name_to_entry(target,
				       value->addr_name_list[i],
				       1 /* strdup it */);
	}
    }

    if ((value->flags & SE_given_name_port) &&
	value->port_name_count > 0) {
	int i;
	target->flags |= SE_given_name_port;

	for (i = 0; i < value->port_name_count; i++) {
	    int j;

	    for (j = 0; j < target->port_name_count; j++) {
		if (0 == istrcmp(value->port_name_list[i],
				 target->port_name_list[j]))
		    break;		
	    }
	    if (j >= target->port_name_count)
		add_port_name_to_entry(target,
				       value->port_name_list[i],
				       1 /* strdup it */);
	}
    }

    if ((value->flags & SE_given_addr) &&
	value->addr_count > 0) {
	int i;
	
	if (! (target->flags & SE_given_addr))
	    free_se_addr_list(target); /* Forget cached addresses */

	target->flags |= SE_given_addr;

	for (i = 0; i < value->addr_count; i++) {
	    int j;

	    for (j = 0; j < target->addr_count; j++) {	    

		if (0 == memcmp(&(value->addr_list[i]),
				&(target->addr_list[j]),
				sizeof (value->addr_list[i])))
		    break;
	    }
	    if (j >= target->addr_count)
		add_addr_to_entry(target,& (value->addr_list[i]));
	}       
    }
    
    if ((value->flags & SE_given_port) &&
	value->port_count > 0) {
	int i;
	
	if (! (target->flags & SE_given_port))
	    free_se_port_list(target);   /* Forget default port */

	target->flags |= SE_given_port;

	for (i = 0; i < value->port_count; i++) {
	    int j;

	    for (j = 0; j < target->port_count; j++) {	    
		if (value->port_list[i] ==
		    target->port_list[j])
		    break;
	    }

	    if (j >= target->port_count)
		add_port_to_entry(target,value->port_list[i]);	    
	}
    }

    if (value->option_count > 0) {
	int i;

	for (i = 0; i < value->option_count; i++) {
	    int j;
	    int idx = -1;
	    
	    for (j = 0; j < target->option_count; j++) {
		if (0 == strcmp(value->option_list[i].prefix,
				target->option_list[j].prefix))
		    idx = j;
	    }
	    if (-1 == idx) {
		struct SE_option_type * Y = 
#ifdef USE_DLOPEN
		    get_option_type(value->option_list[i].prefix)
#else
		    NULL
#endif
		    ;

		if (!Y) {
		    panic("CONNECTION PANIC",__FILE__,__LINE__,
			  "merge1_entries",
			  "option type disappeared",0);
		    continue;
		}
		
		idx = add_option_type_to_entry(target,Y,
					       value->option_list[i].prefix);
					       
	    }
	    if (SE_option_t_magic != target->option_list[idx].type->magic)
		panic("CONNECTION PANIC",__FILE__,__LINE__,
		      "merge1_entries",
		      "Bad option type magic on service list",0);

	    if (target->option_list[idx].type != value->option_list[i].type)
		panic("CONNECTION PANIC",__FILE__,__LINE__,
                      "merge1_entries",
                      "option type mismatch",0);

	    target->option_list[idx].type->
		merge_options(& (target->option_list[idx]),
			      & (value->option_list[i]) );
	}
    }

    if ((value->flags & SE_ver_tls_cert)) {
	target->flags |= SE_ver_tls_cert;

	if ((target->flags & SE_nover_tls_cert)) {
	    target->flags &= ~SE_nover_tls_cert;   
	    DPRINT(Debug,10,(&Debug,  
			     "... conflicting verify-tls-certificate settings, verify-tls-certificate=off ignored\n"));
	}
    }

    if ((value->flags & SE_nover_tls_cert)) {
	if (!(target->flags & SE_ver_tls_cert))
	    target->flags |= SE_nover_tls_cert;
	else {
	    DPRINT(Debug,10,(&Debug,  
			     "... conflicting verify-tls-certificate settings, verify-tls-certificate=off ignored\n"));
	}
    }

    if ((value->flags & SE_require_tls_name)) {
	if (0 == strcmp(target->official_name,value->official_name))	    
	    target->flags |= SE_require_tls_name;
	else {
	    int i;

	    for (i = 0; i < target->req_tls_peername_count; i++) {
		if (string_matches_ascii(target->req_tls_peername_list[i],
					 s2us(value->official_name),
					 0))
		    break;
	    }

	    if (i >= target->req_tls_peername_count) {
		struct string * req = new_string2(system_charset,
						  s2us(value->official_name));
		add_req_peername_to_entry(target,req);
		free_string(&req);
	    }
	}
	
    } else if (value->req_tls_peername_list) {
	int i;

	for (i = 0; i < value->req_tls_peername_count; i++) {
	    add_req_peername_to_entry(target,
				      value->req_tls_peername_list[i]);
	}
    }

}

static void merge_entries P_((struct service_entry **ret,
			      struct service_entry *value));
static void merge_entries(ret,value)
     struct service_entry **ret;
     struct service_entry *value;
{
    if (SERVICE_ENTRY_magic != value->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "merge_entries",
	      "Bad magic (service_entry) (value)",0);

    DPRINT(Debug,12,(&Debug,
		     "merge_entries: value=%p (%s), refcount=%d\n",
		     value,value->official_name,value->refcount));
  
    if (value == *ret)
	return;

    if (!*ret) {
	*ret = value;
	inc_service_entry_refcount(*ret);
    } else { 
	if (SERVICE_ENTRY_magic != (*ret)->magic)
	    panic("CONNECTION PANIC",__FILE__,__LINE__,
		  "merge_entries",
		  "Bad magic (service_entry) (*ret)",0);

	DPRINT(Debug,12,(&Debug,
			 "merge_entries: (*ret)=%p (%s), refcount=%d\n",
			 (*ret),(*ret)->official_name,(*ret)->refcount));


	if ((*ret)->flags & SE_temporary) {
	    merge1_entries(*ret,value);

	} else {
	    struct service_entry * A = 
		malloc_service_entry((*ret)->official_name,
				     (*ret)->service, SE_temporary);
	    merge1_entries(A,*ret);
	    merge1_entries(A,value);

	    free_service_entry(ret);

	    *ret = A;
	    DPRINT(Debug,10,(&Debug,  "... making temporary entry (%p) for %s from permanent\n",
			     (*ret),(*ret)->official_name));
	}
    }

    DPRINT(Debug,12,(&Debug,
		     "merge_entries: result (*ret)=%p (%s), refcount=%d%s\n",
		     (*ret),(*ret)->official_name,(*ret)->refcount,
		     ((*ret)->flags & SE_temporary) ? " (temporary)" : ""));
}

static int name_ok P_((const char *name));
static int name_ok(name)
     const char *name;
{
    if ('\0' == *name)
	return 0;

    return strlen(name) == strspn(name,"abcdefghijklmnopqrstuvxyz.-");
}

static int check_ascii P_((const char *c));
static int check_ascii(c)
     const char *c;
{
    const unsigned char *x;
    
    for (x = cs2us(c); *x; x++) {
	if (!isascii(*x))
	    return 0;
    }

    return 1;
}

/* Mallocs to result */
static char * convert_to_fileset2 P_((struct string *value, 
				      charset_t fileset, int *err));
static char * convert_to_fileset2(value,fileset,err)
     struct string *value;
     charset_t fileset; 
     int *err;
{
    int ERRORS;
    struct string * S2 = NULL;
    unsigned char * res = NULL;	

    S2 = convert_string2(fileset,value,&ERRORS);
    (*err) += ERRORS;

    /* result is malloced */
    res = stream_from_string(S2,0,NULL);
    free_string(&S2);
    
    return us2s(res);
}

/* Mallocs to result, if conversion, otherwise returns value */
static char * convert_to_fileset P_((char *value, charset_t fileset, int *err));
static char * convert_to_fileset(value,fileset,err)
     char *value; 
     charset_t fileset;
     int *err;
{
    if (system_charset != fileset) {
	struct string  * S = new_string(system_charset);

	int ERRORS;
	int len = strlen(value);
	int rs = add_streambytes_to_string(S,len,cs2us(value),&ERRORS);
	char * ret = NULL;

	(*err) += ERRORS;
		
	if (rs < len) {
	    
	    free_string(&S);
	    return NULL;
	}
	
	ret = convert_to_fileset2(S,fileset,err);
	free_string(&S);
	return ret;
    } else
	return value;

}

int dump_service_entry(f,entry,fileset)
     FILE *f; 
     struct service_entry *entry;
     charset_t fileset;
{
    char * sep = "\t";
    int idx;
    int iserr = 0;

    if (SERVICE_ENTRY_magic != entry->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "dump_service_entry",
	      "Bad magic (service_entry)",0);

    if (!check_ascii(entry->official_name) ||
	!check_ascii(entry->service->name))
	return 0;

    fputs(entry->official_name,f);
    putc('\t',f);
    fputs(entry->service->name,f);
	
    if (entry->flags & SE_given_alias) {
	int y;
	for (y = 0; y < entry->aliases_count; y++) {
	    int ERR = 0;
	    
	    char *val = convert_to_fileset(entry->aliases_list[y],
					   fileset,&ERR);
	    iserr += ERR;
	    
	    if (!val)
		continue;
	    
	    if (name_ok(val))
		elm_fprintf(f,FRM("%salias=%s"),
			    sep,val);
	    else
		elm_fprintf(f,FRM("%salias=%Q"),
			    sep,val);
	    sep = "; ";
	    
	    if (val != entry->aliases_list[y])
		free(val);
	}
    }

    if (entry->flags & SE_given_name_addr) {
	int y;
	for (y = 0; y < entry->addr_name_count; y++) {
	    if (!check_ascii(entry->addr_name_list[y])) {
		iserr += 1;
		continue;
	    }

	    if (name_ok(entry->addr_name_list[y]))
		elm_fprintf(f,FRM("%saddr=%s"),
			    sep,entry->addr_name_list[y]);
	    else
		elm_fprintf(f,FRM("%saddr=%Q"),
			    sep,entry->addr_name_list[y]);
	    sep = "; ";
	}
    }

    if (entry->flags & SE_given_name_port) {
	int y;
	for (y = 0; y < entry->port_name_count; y++) {
	    if (!check_ascii(entry->port_name_list[y])) {
		iserr += 1;
		continue;
	    }

	    if (name_ok(entry->port_name_list[y]) &&
		! isdigit(entry->port_name_list[y][0]))
		elm_fprintf(f,FRM("%sport=%s"),
			    sep,entry->port_name_list[y]);
	    else
		elm_fprintf(f,FRM("%sport=%Q"),
			    sep,entry->port_name_list[y]);
	    sep = "; ";
	}
    }
	
    if (entry->flags & SE_given_addr) {
	int y;
	for (y = 0; y < entry->addr_count; y++) {
	    SOCKADDR X = entry->addr_list[y];
	    switch (X.sa.sa_family) {
		char * z;
#ifdef I_NETINET_IN
	    case AF_INET:
		z = inet_ntoa(X.sin.sin_addr);
		elm_fprintf(f,FRM("%saddr=ip:%Q"),
			    sep,z);
		if (ntohs(X.sin.sin_port) != PORT_end)
		    elm_fprintf(f,FRM(":%d"),
				ntohs(X.sin.sin_port));
		
		sep = "; ";
		break;
#endif
	    default:
		/* none */
		break;
	    }		
	}
    }

    if (entry->flags & SE_given_port) {
	int y;
	for (y = 0; y < entry->port_count; y++) {
	    elm_fprintf(f,FRM("%sport=%d"),
			sep,entry->port_list[y]);
	    sep = "; ";
	}
    }

    if (entry->flags & SE_ver_tls_cert) {
	elm_fprintf(f,FRM("%sverify-tls-certificate"),
		    sep);
	sep = "; ";
    }

    if (entry->flags & SE_nover_tls_cert) {
	elm_fprintf(f,FRM("%sverify-tls-certificate=OFF"),
		    sep);
	sep = "; ";
    }

    if (entry->flags & SE_require_tls_name) {
	elm_fprintf(f,FRM("%srequire-tls-peer-name"),
		    sep);
	sep = "; ";
    }

    for (idx = 0; idx < entry->req_tls_peername_count; idx++) {
	int ERR = 0;

	char *val = convert_to_fileset2(entry->req_tls_peername_list[idx],
					   fileset,&ERR);
	iserr += ERR;

	if (!val)
	    continue;
	
	if (name_ok(val))
	    elm_fprintf(f,FRM("%srequire-tls-peer-name=%s"),
			sep,val);
	else
	    elm_fprintf(f,FRM("%srequire-tls-peer-name=%Q"),
			sep,val);
	sep = "; ";
	
	free(val);
    }

    for (idx = 0; idx < entry->option_count; idx++) {
	char * val;
	
	if (SE_option_t_magic != entry->option_list[idx].type->magic)
	    panic("CONNECTION PANIC",__FILE__,__LINE__,
		  "dump_service_entry",
		  "Bad option type magic on service list",0);
	
	val = entry->option_list[idx].type->
	    give_values(& (entry->option_list[idx]),
			entry->option_list[idx].prefix);
	
	if (val) {
	    int ERR = 0;
	    char *val1 = convert_to_fileset(val,fileset,&ERR);
	    iserr += ERR; 
	    
	    if (val1) {
		fputs(sep,f);
		fputs(val1,f);
		sep = "; ";

		if (val1 != val)
		    free(val1);
	    }
	    free(val); val = NULL;
	}
    }
    putc('\n',f);

    if (iserr) {
	const char * sysnam = get_charset_MIME_name(fileset);


	elm_fprintf(f, 
		    CATGETS(elm_msg_cat,MeSet, MeFailedConvertSome,
			    "# Failed to convert some values of %s to %s -charset\n"),
		    entry->official_name,
		    sysnam ? sysnam : "<no MIME name>");
		    
    }

    return !iserr;
}				
 
static int initial_tls_match P_((const struct service_type *st,
				 enum initial_tls_flag initial_tls));
static int initial_tls_match(st,initial_tls)
     const struct service_type *st;
     enum initial_tls_flag initial_tls;
{
    switch (initial_tls) {
    case itls_ignore:
	return 1;
    case itls_disable:
	return NULL == st->have_init_tls;
    case itls_require:
	return NULL != st->have_init_tls;
    }

    return 0; /* Bad value */
}

/* Increments refcount */
static struct service_entry *scan_mail_services_conf 
      P_((struct mail_services_conf *conf,
	  const char *hostname,
	  int flag,
	  enum initial_tls_flag initial_tls));

static struct service_entry *scan_mail_services_conf(conf,hostname,flag,
						     initial_tls)
     struct mail_services_conf *conf;
     const char *hostname;
     int flag;
     enum initial_tls_flag initial_tls;
{
    struct service_entry *ret = NULL;
    int i;

    if (MAIL_SERVICES_magic != conf->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "scan_mail_services_conf",
	      "Bad magic (mail_services_conf) (conf)",0);
    
    for (i = conf->service_count-1; i >= 0; i--) {
	int tls_match = 0;

	if (SERVICE_ENTRY_magic != conf->service_list[i]->magic)
	    panic("CONNECTION PANIC",__FILE__,__LINE__,
		  "scan_mail_services_conf",
		  "Bad magic (service_entry)",0);

	tls_match = initial_tls_match(conf->service_list[i]->service,
				      initial_tls);
	if (0 != (flag & conf->service_list[i]->service->flags) &&
	    tls_match) {
	    int j;

	    if (0 == istrcmp(hostname,conf->service_list[i]->official_name)) {

#ifdef DEBUG
		if (Debug.active >= 8) {
		    FILE *F;
		    DPRINT(Debug,8,(&Debug,"\n-- found [%d] ",i));
		    F = debug_to_FILE(&Debug);
		    if (F) {
			dump_service_entry(F,conf->service_list[i],
					   system_charset);
			fclose(F);
		    }
		}
#endif

		/* ret = service_list[i]	     */
		merge_entries(&ret,conf->service_list[i]);
	    }
	    
	    for (j = 0; j < conf->service_list[i]->aliases_count; j++) {
		if (0 == istrcmp(hostname,
				 conf->service_list[i]->aliases_list[j])) {
#ifdef DEBUG
		    if (Debug.active >= 8) {
			FILE *F;
			DPRINT(Debug,8,(&Debug,"\n-- found [%d] ",i));
			F = debug_to_FILE(&Debug);
			if (F) {
			    dump_service_entry(F,conf->service_list[i], 
					       system_charset);
			    fclose(F);
			}
		    }
#endif

		    /* ret = service_list[i] */
		    merge_entries(&ret,conf->service_list[i]);
		}
	    }
	    
	} else {
	    if (0 == istrcmp(hostname,
			     conf->service_list[i]->official_name)) {
		DPRINT(Debug,8,(&Debug,
				"\nscan_mail_services_conf(%s,%d): [%d] official_name=%s  %s flag=%d MISMATCH%s\n",
				hostname,flag,i,
				conf->service_list[i]->official_name,
				conf->service_list[i]->service->name,
				conf->service_list[i]->service->flags,
				tls_match ? ", tls ok" : " on tls"));
	    }
	}	
    }

    return ret;
}
  

/* Increments refcount */
static struct service_entry *scan_list P_((const char *hostname,
					   int flag,
					   enum initial_tls_flag initial_tls));
static struct service_entry *scan_list(hostname,flag,initial_tls)
     const char *hostname;
     int flag;
     enum initial_tls_flag initial_tls;
{
    struct service_entry *ret = NULL;

    if (user_mail_services_conf) {
	ret = scan_mail_services_conf(user_mail_services_conf,hostname,flag,
				      initial_tls);
	
	if (ret) {
	    DPRINT(Debug,12,(&Debug,
			     "scan_list(%s,%d): %p -- Found from user_mail_services_conf\n",
			     hostname,flag,ret));
	}


	if (system_mail_services_conf) {
	    struct service_entry *ret1 = 
		scan_mail_services_conf(system_mail_services_conf,hostname,
					flag,initial_tls);

	    if (ret1) {

		DPRINT(Debug,12,(&Debug,
				 "scan_list(%s,%d): %p -- Found from system_mail_services_conf\n",
				 hostname,flag,ret1));


		merge_entries(&ret,ret1);
		free_service_entry(&ret1);
	    }
	}

    } else if (system_mail_services_conf) {
	ret = scan_mail_services_conf(system_mail_services_conf,hostname,
				      flag,initial_tls);

	if (ret) {
	    DPRINT(Debug,12,(&Debug,
			     "scan_list(%s,%d): %p -- Found from system_mail_services_conf\n",
			     hostname,flag,ret));
	}
	
    } else {
	DPRINT(Debug,8,(&Debug,
			"scan_list: No user_mail_services_conf or system_mail_services_conf\n"));
    }

#ifdef DEBUG
    if (Debug.active >= 8 && ret) {
	FILE *F;
	if (ret->flags & SE_temporary) {
	    DPRINT(Debug,8,(&Debug,"\nMERGED        "));
	} else {
	    DPRINT(Debug,8,(&Debug,"\nRESULT        "));
	}
	F = debug_to_FILE(&Debug);
	if (F) {
	    dump_service_entry(F,ret, system_charset);
	    fclose(F);
	}
	DPRINT(Debug,8,(&Debug,"\n"));
    }
#endif

    if (ret) {
    	DPRINT(Debug,12,(&Debug,"scan_list(%s,%d)=%p, %s: refcount=%d\n",
			 hostname,flag,ret,ret->official_name,ret->refcount));
    } else {
	DPRINT(Debug,12,(&Debug,"scan_list(%s,%d)=NULL\n",
			 hostname,flag));
    }

    return ret;
}

static struct hostent * lookup_name P_((const char *hostname,int silent));
static struct hostent * lookup_name(hostname,silent)
     const char *hostname;
     int silent;
{
    struct hostent *he = NULL;

    const char * reserved_name = is_special_use_domain(hostname);

    if (reserved_name) {

	if (!silent)
	    lib_error(CATGETS(elm_msg_cat, MeSet,
			      MeReservedDomainUnsupported,
			      "Reserved domain %s is unsupported: %s"),
		      reserved_name,hostname);

	DPRINT(Debug,9,(&Debug, 
			"lookup_name: %s: special use domain %s\n",
			hostname,reserved_name));

	return NULL;
    }


    lib_transient(CATGETS(elm_msg_cat, MeSet,MeLookingUp,
			  "Looking up %s ..."),
		  hostname);
	
    he = gethostbyname(hostname);
	
    if (!he) {
	DPRINT(Debug,9,(&Debug, 
			"lookup_name: %s: h_errno=%d\n",
			hostname,h_errno));
	    
	/* We are silent about errors if cached
	   data exists
	*/
	if (!silent) {
	    switch(h_errno) {
	    case HOST_NOT_FOUND:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeNoRemoteHost,
				  "No remote server host: %s"),
			  hostname);		
		break;
	    case NO_ADDRESS:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeNoAddress,
				  "Name %s have not IP-address"),
			  hostname);
		break;
	    case TRY_AGAIN:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeTemporary,
				  "Address for %s is not yet found..."),
			  hostname);
		break;			  
	    default:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeNameError,
				  "Failed to get address for %s"),
			  hostname);
		break;
	    }
	}
    } else {
	lib_transient(CATGETS(elm_msg_cat, MeSet,MeLookingUpOK,
			      "Looking up %s ... OK"),
		      hostname);

    }

    return he;
}

static void fill_port P_((struct service_entry *ret,
			  struct servent  *se));
static void fill_port(ret,se)
     struct service_entry *ret;
     struct servent  *se;
{   
    if (SERVICE_ENTRY_magic != ret->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "fill_port",
	      "Bad magic (service_entry)",0);

    add_port_to_entry(ret,ntohs(se->s_port));
}


static void fill_address P_((struct service_entry *ret,
			     struct hostent *he));
static void fill_address(ret,he)
     struct service_entry *ret;
     struct hostent *he;
{   
    if (SERVICE_ENTRY_magic != ret->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "fill_address",
	      "Bad magic (service_entry)",0);

    if (he->h_addr_list) {
	int x;
	for (x = 0; he->h_addr_list[x]; x++) {
	    SOCKADDR X;

	    /* Make comparision easier */
	    bzero((void *)&X, sizeof X);

	    switch(he->h_addrtype) {
#ifdef I_NETINET_IN
	    case AF_INET:
		X.sin.sin_family = he->h_addrtype;
	        X.sin.sin_port = htons(PORT_end);    /* No port given */

		if (sizeof(X.sin.sin_addr)  != he->h_length) {
		    lib_error(FRM("%s: Bad IP-addr length %d (should be %d)"),
			      he->h_name,he->h_length,
			      sizeof(X.sin.sin_addr));
		    return;
		}
		memcpy(&(X.sin.sin_addr), he->h_addr_list[x], he->h_length);
		add_addr_to_entry(ret,&X);
		break;
#endif			    
	    default:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeUnsupportedAddrType,
				  "Name %s have odd type address"),
			  he->h_name);
		/* NOTE: We do not know on what components arbitary
		         socket address is constructed so we can't
			 fill it
		*/
		return;
	    }
	}
    }
}

static int  addrs_have_port P_((struct service_entry *se));
static int  addrs_have_port(se)
     struct service_entry *se;
{
    int have_port = 0;

    if (SERVICE_ENTRY_magic != se->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "addrs_have_port",
	      "Bad magic (service_entry)",0);


    if (se->addr_count > 0) {
	int x;

	have_port = 1;

	for (x = 0; x < se->addr_count; x++) {
	    switch(se->addr_list[x].sa.sa_family) {
#ifdef I_NETINET_IN
	    case AF_INET:
		if (se->addr_list[x].sin.sin_port == htons(PORT_end))
		    have_port = 0;
		break;
#endif
	    default:
		have_port = 0;
		break;
	    }
	}

    }

    DPRINT(Debug,14,(&Debug,
		     "addrs_have_port=%d, se=%p; %d addresses\n",
		     have_port,se,se->addr_count));

    return have_port;
}

/* Increments refcount */
struct service_entry * give_service_entry2(hostname,flag,initial_tls)
     const char          * hostname;
     int                   flag;
     enum initial_tls_flag initial_tls;
{
    struct service_entry *ret;       
    int have_port = 0;   /* set if addresses have port set */
    int was_error = 0;

    DPRINT(Debug,9,(&Debug, 
		    "give_service_entry(name=\"%s\",flags=%d, initial_tls=%d)\n",
		    hostname,flag,initial_tls));

    ret = scan_list(hostname,flag,initial_tls);
    /* If on service entry do not have ip addresses
       they needed to be retrieved. Notice that we
       retrieve new entries every time (alternatively
       we should keep track of expiry times).
       However all other data is cached (also addresses if
       them is generated from names given on configuration data)
    */
    if (ret)
	hostname = ret->official_name;

    if (ret && (
		( 0 != (ret->flags & SE_given_name_addr) &&
		  ret->addr_count == 0 ) 
		||
		0 != (ret->flags & SE_rescan_name_addr)
		)
	) {
	int x;

	/* Clear data on case of rescan */
	if (ret->addr_list) {
	    free(ret->addr_list);
	    ret->addr_list = NULL;
	    ret->addr_count = 0;
	}
	ret->flags   &= ~SE_rescan_name_addr;

	for (x = 0; x < ret->addr_name_count; x++) {
	    struct hostent *he  = 
		lookup_name(ret->addr_name_list[x],0);
	    
	    if (he)
		fill_address(ret,he);
	    else {
		ret->flags |= SE_rescan_name_addr;
		was_error = 1;
	    }
	}
    }

    if (!ret || (
		 0 == (ret->flags & SE_given_addr) &&
		 0 == (ret->flags & SE_given_name_addr)
		 )
	) {

	struct hostent *he    = lookup_name(hostname,
					    /* Silent: */
					    ret && ret->addr_count > 0);

	if (he) {  	  
	    /* Now try found entry again ... */
	    if (!ret) { 
		ret = scan_list(he->h_name,flag,initial_tls);
		
		if (!ret && he->h_aliases) {
		    int j;
		    for (j = 0; he->h_aliases[j] && !ret; j++)
			ret = scan_list(he->h_aliases[j],flag,
					initial_tls);
		}
	    
		/* we need to add hostname to aliases because
		   it was not on there (otherwise entry
		   should have found earlier)
		*/
		if (ret) 
		    add_alias_to_entry(ret,hostname);
	    }
	    
	    if (!ret) {
		struct service_type *st;

		for (st = &SERVICE_TYPES[0]; st->name; st++) {
		    if (flag == (flag &st->flags) &&
			initial_tls_match(st,initial_tls))
			break;
		}

		if (!st->name)
		    panic("CONNECTION PANIC",__FILE__,__LINE__,
			  "give_service_entry",
			  "Service type do not found", 0);

		ret = malloc_service_entry(hostname,st,SE_temporary);
	    }
	
	    if (he->h_aliases) {  /* Add missing aliases */
		int x;
		for (x = 0; he->h_aliases[x]; x++) {
		    int j;
		    for (j=0; j < ret->aliases_count; j++)
			if (0 == istrcmp(he->h_aliases[x],
					 ret->aliases_list[j]))
			    break;
		    if (j >= ret->aliases_count) 
			add_alias_to_entry(ret,he->h_aliases[x]);
		}
	    }

	    if (he->h_addr_list) {
		/* Clear old data */
	    	if (ret->addr_list) {
		    free(ret->addr_list);
		    ret->addr_list = NULL;
		    ret->addr_count = 0;
		}

		fill_address(ret,he);
	    }		
	}    
    }

    if (ret && ret->addr_count > 0)
	have_port = addrs_have_port(ret);


    if (ret && !have_port && (
			      ( 0 != (ret->flags & SE_given_name_port) &&
				ret->port_count == 0 ) 
			      ||
			      0 != (ret->flags & SE_rescan_name_port)
			      )
	) { 

	int x;
	int was_rescan = 0 != (ret->flags & SE_rescan_name_port);
	
	if (ret->port_list) {
	    free(ret->port_list);
	    ret->port_list = NULL;
	    ret->port_count = 0;
	}
	ret->flags   &= ~SE_rescan_name_port;

	for (x = 0; x < ret->port_name_count; x++) {
	    struct servent * se =
		getservbyname(ret->port_name_list[x],
			      "tcp");

	    if (!se && !was_rescan) {
		lib_error(CATGETS(elm_msg_cat, MeSet,
				  MeNoTcpPortNotFound,
				  "%s: TCP port name %s not found"),
			  ret->official_name,
			  ret->port_name_list[x]);
		was_error = 1;
	    }

	    if (se)
		fill_port(ret,se);
	    else
		ret->flags |= SE_rescan_name_port;
	}
    }

    if (ret && ret->port_count == 0 && !have_port &&
	0 == (ret->flags & SE_rescan_name_port)) {
	PORTS r = PORT_end;
	
	/* builtin definations here */
	if ( PORT_end == ret->service->defport) {
	    
	    /* Caller will define portlist */
	    if (ret->port_list)
	 	free(ret->port_list);
	    ret->port_list = NULL;
	    
	} else 
	    r = ret->service->defport;
	
	if (r != PORT_end) {
	    ret->port_list = safe_realloc(ret->port_list,
					  2 * sizeof (ret->port_list[0]));
	    ret->port_list[0] = r;
	    ret->port_list[1] = PORT_end;
	    ret->port_count  = 1;
	}
    }

    if (was_error) {
	DPRINT(Debug,1,(&Debug, 
			"give_service_entry: %s: Was error\n",
			hostname));

	if (sleepmsg > 0) {  /* HACK */
#if POLL_METHOD   
	    wait_for_timeout(sleepmsg);
#else
	    sleep(sleepmsg);
#endif
	}

    }

    if (ret) {
	DPRINT(Debug,9,(&Debug, 
			"give_service_entry=%p: type=%p (%s)%s, refcount=%d\n",
			ret,ret->service,ret->service->name,
			ret->flags & SE_temporary ? ", temporary" : "",
			ret->refcount));

    } else {
	DPRINT(Debug,9,(&Debug, 
			"give_service_entry=NULL\n"));
    }

    return ret;
}

/* Increments refcount */
struct service_entry * give_service_entry(hostname,flag)
     const char *hostname;
     int flag;
{
    return give_service_entry2(hostname,flag,itls_ignore);
}



void free_se_aliases_list(se) 
     struct service_entry * se;
{
    if (SERVICE_ENTRY_magic != se->magic)
	 panic("CONNECTION PANIC",__FILE__,__LINE__,
	       "free_se_aliases_list",
	       "Bad magic (service_entry) (*ret)",0);

    if (se->aliases_list) {
	int x;
		
	for (x = 0; x < se->aliases_count; x++) {
	    if (se->aliases_list[x])
		free(se->aliases_list[x]);
	    se->aliases_list[x] = NULL;
	}
	
	free(se->aliases_list);
	se->aliases_list  = NULL;
    }
    se->aliases_count = 0;    
}

void free_se_addr_list(se) 
     struct service_entry * se;
{
    if (SERVICE_ENTRY_magic != se->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "free_se_addr_list",
	      "Bad magic (service_entry) (*ret)",0);
    
    if (se->addr_list) {
	free(se->addr_list);
	se->addr_list = NULL;
    }
    se->addr_count = 0;
}

void free_se_port_name_list(se) 
     struct service_entry * se;
{
    if (SERVICE_ENTRY_magic != se->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "free_se_port_name_list",
	      "Bad magic (service_entry) (*ret)",0);

    if (se->port_name_list) {
	int x;
	
	for (x = 0; x < se->port_name_count; x++) {
	    if (se->port_name_list[x])
		free(se->port_name_list[x]);
	    se->port_name_list[x] = NULL;
	}
	
	free(se->port_name_list);
	se->port_name_list = NULL;
    }
    se->port_name_count = 0;
}
    
void free_se_addr_name_list(se) 
     struct service_entry * se;
{
    if (SERVICE_ENTRY_magic != se->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "free_se_addr_name_list",
	      "Bad magic (service_entry) (*ret)",0);

    if (se->addr_name_list) {
	int x;

	for (x = 0; x < se->addr_name_count; x++) {
	    if (se->addr_name_list[x])
		free(se->addr_name_list[x]);
	    se->addr_name_list[x] = NULL;
	}
	
	free(se->addr_name_list);
	se->addr_name_list = NULL;
    }
    se->addr_name_count = 0;
}

void free_se_port_list(se) 
     struct service_entry * se;
{
    if (SERVICE_ENTRY_magic != se->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "free_se_addr_name_list",
	      "Bad magic (service_entry) (*ret)",0);

    if (se->port_list)
	free(se->port_list);
    se->port_list  = NULL;
    se->port_count = 0;
}

static void free_se_option_list P_((struct service_entry * se));
static void free_se_option_list(se) 
     struct service_entry * se;
{
    if (SERVICE_ENTRY_magic != se->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "free_se_option_list",
	      "Bad magic (service_entry) (*ret)",0);

    if (se->option_list) {
	int x;
	
	for (x = 0; x < se->option_count; x++) {
	    if (SE_option_t_magic != se->option_list[x].type->magic)
		panic("CONNECTION PANIC",__FILE__,__LINE__,
		      "free_se_option_list",
		      "Bad option type magic on service list",0);
	    
	    se->option_list[x].type->
		free_options(& (se->option_list[x]));	    
	    
	    if (se->option_list[x].prefix)
		free(se->option_list[x].prefix);
	    se->option_list[x].prefix = NULL;
	}

	free(se->option_list);
	se->option_list = NULL;
    }
    se->option_count = 0;
}

void free_se_rqtls_pername_list(se) 
     struct service_entry * se;
{
    if (SERVICE_ENTRY_magic != se->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "free_se_rqtls_pername_list",
	      "Bad magic (service_entry) (*ret)",0);

    if (se->req_tls_peername_list) {
	int x;

	for (x = 0; x < se->req_tls_peername_count; x++)
	    if (se->req_tls_peername_list[x])
		free_string(& (se->req_tls_peername_list[x]));
	free(se->req_tls_peername_list);
	se->req_tls_peername_list = NULL;
    }
    se->req_tls_peername_count = 0;
}

void inc_service_entry_refcount(Y)
      struct service_entry *Y;
 {
     if (SERVICE_ENTRY_magic != Y->magic)
	 panic("CONNECTION PANIC",__FILE__,__LINE__,
	       "inc_service_entry_refcount",
	       "Bad magic (service_entry) (*ret)",0);

     Y->refcount++;
 }

void free_service_entry(Y)
     struct service_entry **Y;
{
    if (!(*Y))
	return;

    if (SERVICE_ENTRY_magic != (*Y)->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "free_service_entry",
	      "Bad magic (service_entry)",0);
    
    DPRINT(Debug,12,(&Debug, 
		     "free_service_entry: entry %p: %s%s\n",
		     (*Y), (*Y)->official_name,  
		     ((*Y)->flags & SE_temporary) ? ", (temporary)" : ""));
	   
    (*Y)->refcount--;

    if ((*Y)->refcount > 0) {

	DPRINT(Debug,12,(&Debug, 
			 "free_service_entry: entry %p: refcount=%d\n",
			 *Y,(*Y)->refcount));

	*Y = NULL;
	return;
    }

    DPRINT(Debug,12,(&Debug, 
		     "free_service_entry: Freeing entry %p\n",
		     *Y));

    if ((*Y)->flags  & SE_temporary) {
	DPRINT(Debug,12,(&Debug, 
			 "free_service_entry: Is temporary\n"));
    }

    if ((*Y)->official_name)
	free((*Y)->official_name);
    (*Y)->official_name = NULL;


    free_se_aliases_list(*Y);
    free_se_addr_list(*Y);
    free_se_addr_name_list(*Y);
    free_se_port_name_list(*Y);
    free_se_port_list(*Y);

    (*Y)->service = NULL;

    free_se_option_list(*Y);
    free_se_rqtls_pername_list(*Y);

    (*Y)->magic = 0;  /* Invalidate */

    free(*Y);
    (*Y) = NULL;
}

int is_similar_service_entry(se1,se2)
     struct service_entry * se1;
     struct service_entry * se2;
{
    int ret = 0;

    if (SERVICE_ENTRY_magic != se1->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "is_similar_service_entry",
	      "Bad magic (service_entry) (se1)",0);

    if (SERVICE_ENTRY_magic != se2->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "is_similar_service_entry",
	      "Bad magic (service_entry) (se2)",0);

    if (se1->service != se2->service) {
	DPRINT(Debug,12,(&Debug,
			 "is_similar_service_entry: service differ\n"));
    }

    if (0 != strcmp(se1->official_name,se2->official_name)) {
	DPRINT(Debug,12,(&Debug,
			 "is_similar_service_entry: name differ\n"));
	goto fail;
    }

    if (se1->flags != se2->flags) {
	DPRINT(Debug,12,(&Debug,
			 "is_similar_service_entry: flags differ\n"));
	goto fail;	

    } else if (se1->flags) {
	
	/* Compare only flagged data, this does not compare
	   generated (hidden) data
	*/
	
	if (0 != (se1->flags & SE_given_alias)) {
	    int y;
	    
	    if (se1->aliases_count != se2->aliases_count) {
		DPRINT(Debug,12,(&Debug,
				 "is_similar_service_entry: aliases differ\n"));
		goto fail;	
	    }
	    
	    for (y = 0; y < se1->aliases_count; y++) {
		if (0 != strcmp(se1->aliases_list[y],
				se2->aliases_list[y])) {
		    DPRINT(Debug,12,(&Debug,
				     "is_similar_service_entry: aliases differ\n"));
		    goto fail;
		}
	    }
	    
	    DPRINT(Debug,13,(&Debug,
			     "is_similar_service_entry: aliases (%d) are same\n",
			     y));

	} 
	
	if (0 != (se1->flags & SE_given_name_addr)) {
	    int y;
	    
	    if (se1->addr_name_count != se2->addr_name_count) {
		DPRINT(Debug,12,(&Debug,
				 "is_similar_service_entry: named addr differ\n"));
		goto fail;
	    }
	    
	    for (y = 0; y < se1->addr_name_count; y++) {
		if (0 != strcmp(se1->addr_name_list[y],
				se2->addr_name_list[y])) {
		    DPRINT(Debug,12,(&Debug,
				     "is_similar_service_entry: named addr differ\n"));
		    goto fail;
		}
	    }
	    
	    DPRINT(Debug,13,(&Debug,
			     "is_similar_service_entry: named addrs (%d) are same\n",
			     y));
	    
	} 
	
	if (0 != (se1->flags & SE_given_name_port)) {
	    int y;
	    
	    if (se1->port_name_count != se2->port_name_count) {
		DPRINT(Debug,12,(&Debug,
				 "is_similar_service_entry: named port differ\n"));
		goto fail;
	    }
	    
	    for (y = 0; y < se1->port_name_count; y++) {
		if (0 != strcmp(se1->port_name_list[y],
				se2->port_name_list[y])) {
		    DPRINT(Debug,12,(&Debug,
				     "is_similar_service_entry: named port differ\n"));
		    goto fail;
		}
	}
	    
	DPRINT(Debug,13,(&Debug,
			 "is_similar_service_entry: named ports (%d) are same\n",
			 y));
	
	} 
	
	if (0 != (se1->flags & SE_given_addr)) {
	    int y;
	    
	    if (se1->addr_count != se2->addr_count) {
		DPRINT(Debug,12,(&Debug,
				 "is_similar_service_entry: addr differ\n"));
		goto fail;
	    }
	    
	    for (y = 0; y < se1->addr_count; y++) {
		
		if (0 != memcmp(&(se1->addr_list[y]),
				&(se2->addr_list[y]),
				sizeof (se1->addr_list[y]))) {
		    DPRINT(Debug,12,(&Debug,
				     "is_similar_service_entry: addr differ\n"));
		    goto fail;
		}
	    }
	    
	    DPRINT(Debug,13,(&Debug,
			     "is_similar_service_entry: addrs (%d) are same\n",
			     y));
	    
	} 

	if (0 != (se1->flags & SE_given_port)) {
	    int y;
	    
	    if (se1->port_count != se2->port_count) {
		DPRINT(Debug,12,(&Debug,
				 "is_similar_service_entry: port differ\n"));
		goto fail;
	    }
	    
	    for (y = 0; y < se1->port_count; y++) {
		if (se1->port_list[y] != se2->port_list[y]) {
		    DPRINT(Debug,12,(&Debug,
				     "is_similar_service_entry: port differ\n"));
		    goto fail;
		}
	    }	
	    
	    DPRINT(Debug,13,(&Debug,
			     "is_similar_service_entry: ports (%d) are same\n",
			     y));
	    
	} 
    }	

    if (se1->req_tls_peername_count != se2->req_tls_peername_count) {
	DPRINT(Debug,12,(&Debug,
			 "is_similar_service_entry: req tls peername differ\n"));
	goto fail;
    } else if (se1->req_tls_peername_count) {
	int y;

	for (y = 0; y < se1->req_tls_peername_count; y++) {
	    if (0 != string_cmp(se1->req_tls_peername_list[y],
				se2->req_tls_peername_list[y],
				-99 /* Unknown value*/)) {
		DPRINT(Debug,12,(&Debug,
				 "is_similar_service_entry: req tls peername differ\n"));
		goto fail;
	    } 
	}

	DPRINT(Debug,13,(&Debug,
			 "is_similar_service_entry: req tls peernames (%d) are same\n",
			 y));
	
    }
    
    if (se1->option_count != se2->option_count) {
	DPRINT(Debug,12,(&Debug,
			 "is_similar_service_entry: options differ\n"));
	goto fail;
    } else if (se1->option_count) {
	int idx;

	for (idx = 0; idx < se1->option_count; idx++) {
	    char * val1, * val2;;

	    if (SE_option_t_magic != se1->option_list[idx].type->magic)
		panic("CONNECTION PANIC",__FILE__,__LINE__,
		      "is_similar_service_entry",
		      "Bad option type magic on service list",0);

	    if (se1->option_list[idx].type !=
		se2->option_list[idx].type) {
		DPRINT(Debug,12,(&Debug,
				 "is_similar_service_entry: options differ\n"));
		goto fail;
	    }

	    val1 = se1->option_list[idx].type->
		give_values(& (se1->option_list[idx]),
			    se1->option_list[idx].prefix);
	    val2 = se2->option_list[idx].type->
		give_values(& (se2->option_list[idx]),
			    se2->option_list[idx].prefix);

	    if (val1 && val2) {
		if (0 != strcmp(val1,val2)) {

		    free(val1);
		    free(val2);

		    DPRINT(Debug,12,(&Debug,
				     "is_similar_service_entry: options differ\n"));
		    goto fail;

		}
	    } else if (val1 || val2) {

		if (val1)
		    free(val1);
		if (val2)
		    free(val2);
		
		DPRINT(Debug,12,(&Debug,
				 "is_similar_service_entry: options differ\n"));
		goto fail;
	    }

	    if (val1)
		free(val1);
	    if (val2)
		free(val2);
	    
	}

	DPRINT(Debug,13,(&Debug,
			 "is_similar_service_entry: options (%d extensions) are same\n",
			 idx));
	
    }

    ret = 1;

 fail:
#ifdef DEBUG
    DPRINT(Debug,12,(&Debug,"is_similar_service_entry=%d",
		     ret));
    
    if (Debug.active >= 13) {
	FILE *F;
		
	DPRINT(Debug,13,(&Debug,"\n  -- se1=%p : ",se1));

	F = debug_to_FILE(&Debug);
	dump_service_entry(F,se1,system_charset);
	fclose(F);

	DPRINT(Debug,13,(&Debug,"\n  -- se2=%p : ",se2));

	F = debug_to_FILE(&Debug);
	dump_service_entry(F,se2,system_charset);
	fclose(F);
    }
    DPRINT(Debug,12,(&Debug,"\n"));
#endif

    return ret;
}

static struct string *convert_se_value P_((const char *filename,
					   int lineno,
					   int *errors,
					   const char *opt,
					   charset_t cs,
					   const char *val,
					   int vallen));
static struct string *convert_se_value(filename,lineno,errors,opt,cs,val,vallen)
     const char *filename;
     int lineno;
     int *errors;
     const char *opt;
     charset_t cs;
     const char *val;
     int vallen;
{
    struct string * buffer = new_string(cs);
    const char * csn  = get_charset_MIME_name(cs);

    int ERRORS;
    int rs = add_streambytes_to_string(buffer,vallen,cs2us(val),&ERRORS);

    if (rs < vallen) {
	lib_error(CATGETS(elm_msg_cat, MeSet, MeOptionValueCharset,
			  "%s: %d: Failed to parse option \"%s\" value as charset %s\n"),
		  filename, lineno, opt, csn ? csn :  "<no MIME name>");
	(*errors) ++;

	free_string(&buffer);
	goto fail;
    }

    if (ERRORS) {
	lib_error(CATGETS(elm_msg_cat, MeSet, MeOptionValueCSErrors,
			  "%s: %d: There is %d errors when parsing option \"%s\" value as charset %s"),	  
		  filename, lineno,ERRORS,opt, csn ? csn :  "<no MIME name>");

	(*errors) ++;
    }

 fail:
    return buffer;
}

static int is_ascii_value P_((const char *filename,
			      int lineno,
			      int *errors,
			      const char *opt,
			      const char *val));
static int is_ascii_value(filename,lineno,errors,opt,val)
     const char *filename;
     int lineno;
     int *errors;
     const char *opt;
     const char *val;
{

    const unsigned char *x;

    for (x = cs2us(val); *x; x++) {
	 if (!isascii(*x)) {
	     if (opt)
		 lib_error(CATGETS(elm_msg_cat, MeSet, MeAsciiOptionValue,
				   "%s: %d: Only ASCII characters (not 0x%02x) are allowed on  option \"%s\" value"),	  
			   filename, lineno, *x,opt);
	     else
		 lib_error(CATGETS(elm_msg_cat, MeSet, MeAsciiFieldValue,
				   "%s: %d: Only ASCII characters (not 0x%02x) are allowed on field"),
			   filename, lineno, *x);

	     (*errors) ++;
	     return 0;
	 }

    }

    return 1;
}

static int rc_value_to_syscs P_((const char *filename,
				 int lineno,
				 int *errors,
				 const char *opt,
				 const struct string * buffer,
				 char **result));
static int rc_value_to_syscs(filename,lineno,errors,opt,buffer,result)
     const char *filename;
     int lineno;
     int *errors;
     const char *opt;
     const struct string * buffer;
     char **result;
{
    const char *csn = get_string_MIME_name(buffer);
    const char * rcsetnam = get_charset_MIME_name(system_charset);
    int ret = 0;

    int ERRORS;
    struct string * res_S = convert_string2(system_charset,buffer,&ERRORS);

    if (*result) {
	free(*result);
	*result = NULL;
    }

    if (ERRORS) {
	lib_error(CATGETS(elm_msg_cat, MeSet, MeConvOptionValErrors,
			  "%s: %d: There is %d errors when converting option \"%s\" value  from charset %s to charset %s"),
		  filename,lineno,ERRORS,opt,
		  csn ? csn : "<no MIME name>",
		  rcsetnam ? rcsetnam : "<no MIME name>");

	(*errors) ++;
    }

     /* result is malloced */
    bytestream_from_string(res_S,result,&ret);

    return ret;
}
				 
void free_mail_services_conf(conf)
     struct mail_services_conf **conf;
{
    if (MAIL_SERVICES_magic != (*conf)->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "free_mail_services_conf",
	      "Bad magic ( mail_services_conf)",0);
 
    if ((*conf)->service_list) {
	int i; 

	for (i = 0; i < (*conf)->service_count; i++) {
	    free_service_entry(& ((*conf)->service_list[i]));
	}

	free((*conf)->service_list);
	(*conf)->service_list = NULL;
    }
    (*conf)->service_count = 0;

    (*conf)->magic = 0;   /* Invalidate */

    free(*conf);
    *conf = NULL;
}

struct mail_services_conf * malloc_mail_services_conf()
{
    struct mail_services_conf *ret = safe_malloc(sizeof (*ret));

    /* bzero is defined hdrs/elm_defs.h */
    bzero((void *)ret,sizeof (*ret));

    ret->service_count  = 0;
    ret->service_list   = NULL;
    ret->magic          = MAIL_SERVICES_magic;

    return ret;
}


struct mail_services_conf * parse_service_entries (filename,errors, 
						   fileset, propline)
     const char *filename; 
     int *errors;
     charset_t *fileset;
     struct editor_propline **propline;
{
    struct mail_services_conf *ret = NULL;

    /*   hostname   service        options               */
    int max_count = 0;
    FILE * f;
    char * buf = NULL;
    int c;
    charset_t cs = system_charset;
    int lineno = 0;
    int last_c = 0;
    int err = can_open(filename,"r");

    if (err) {
	DPRINT(Debug,2,(&Debug, 
			"parse_service_entries=NULL: %s: %s (can_open)\n",
			filename,strerror(err)));
	return NULL;
    }

    f = fopen(filename,"r");
    if (!f) {
	int err UNUSED_VAROK = errno;
	DPRINT(Debug,2,(&Debug, 
			"parse_service_entries=NULL: %s: %s\n",
			filename,strerror(err)));
	return NULL;
    }

    if (fileset && *fileset) {
	const char * cs_name = get_charset_MIME_name(*fileset);
	
	cs = *fileset;

	if (cs_name) {
	    DPRINT(Debug,2,(&Debug, 
			    "parse_service_entries: %s: default charset: %s\n",
			    filename,cs_name));
	}
    }

    /* Look for editor property line */

    if (propline)  {
	enum editor_propline_v propline_mode = 
	    give_dt_enumerate_as_int(&editor_ms_propline);

	/* Nothing detected yet */

	if (*propline)
	    free_editor_propline(propline);

	if (propline_mode != editor_propline_ignore) {
	    int i;
	    
	    for (i = 0; i < 3 && !*propline; i++) {
		int l1;

		c = fgetc(f);

		if (EOF == c)
		    goto propline_failure;
		if ('\n' == c)
		    continue;

		ungetc(c,f);

		if ('#' != c)
		    break;

		l1 = malloc_gets(&buf,LONG_STRING,f);
		if (l1 > 1) {

		    if ('#' != buf[0])
			goto propline_failure;
		    
		    *propline =
			detect_editor_propline(propline_mode,
					       buf,l1,filename,i,&cs);
					       
		} else { /* something wrong? */
		propline_failure:

		    rewind(f);
		    break;
		}
	    }
	}	
    }

    while(EOF != (c = fgetc(f))) {
	if ('\n' == c)
	    max_count++;
	last_c = c;
    }

    if (last_c && '\n' != last_c) {
	DPRINT(Debug,9,(&Debug, 
			"parse_service_entries: %s, no newline on end of file\n",
			filename));
	max_count++;	
    }

    DPRINT(Debug,9,(&Debug, 
		    "parse_service_entries: %s, max_count=%d\n",
		    filename,max_count));

    ret = malloc_mail_services_conf();
    
    if (!max_count) {
	fclose(f);
	return ret;
    }
    rewind(f);

    ret->service_list =
	safe_malloc(max_count * sizeof (ret->service_list[0]));

    while (ret->service_count < max_count) {
	
	int l1 = malloc_gets(&buf,LONG_STRING,f);

	char *c,*c1, *options;
	struct service_type *st;
	struct service_entry * se = NULL;

	if (-1 == l1) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeTooLongLineNo,
			      "%s: %d: Too long line: %.30s..."),
		      filename,lineno+1,buf);
	    (*errors) ++;
	    break;
	} else if (l1 < 0) {
	    DPRINT(Debug,9,(&Debug, 
			    "parse_service_entries: %s: %d: read error or EOF\n",
			    filename,lineno+1));
	    break;
	}

	lineno++;

	if (l1 == 0)
	    continue;

	while (l1-- > 0 && whitespace(buf[l1]))
	    buf[l1] = '\0';
	
	if (l1 > 8 && 0 == strncmp(buf,"@charset",8)) {
	    charset_t A;
	    
	    c = buf +8;
	    
	    while (*c && whitespace (*c)) /* skip whitespace */
		c++;

	    if ('=' != *c) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeBadCharsetLine,
				  "%s: %d: Bad @charset line: %s"),
			  filename,lineno, buf);
		(*errors) ++;
		break;	  		
	    }
	    c++;

	    while (*c && whitespace (*c)) /* skip whitespace */
		c++;
	    
	    A = MIME_name_to_charset(c,0);
	
	    if (A) {
		const char * csn = get_charset_MIME_name(A);

		if (!(CS_mapping & charset_properties(A))) {
		    lib_error(CATGETS(elm_msg_cat, MeSet, 
				      MeUnsupportedCharsetLine,
				      "%s: %d: Unsupported @charset '%s' on line: %s"),
			      filename,lineno,csn ? csn : c,buf);
		    (*errors) ++;
		    continue;
		}

		cs = A;

		if (csn) {
		    DPRINT(Debug,2,(&Debug, 
				    "parse_service_entries: %s: charset: %s (%s)\n",
				    filename,c,csn));

		} else {
		    DPRINT(Debug,2,(&Debug, 
				    "parse_service_entries: %s: charset: %s\n",
				    filename,c));
		}
	    		
	    } else {
		lib_error(CATGETS(elm_msg_cat, MeSet, 
				  MeUnsupportedCharsetLine,
				  "%s: %d: Unsupported @charset '%s' on line: %s"),
			  filename,lineno,c,buf);
		(*errors) ++;
	    }

	    continue;
	}
	    
	c = buf;
	while (*c && whitespace (*c)) /* skip leading whitespace */
	    c++;
	if ('#' == *c)
	    continue;
	if (!*c)
	    continue;
	
	c1 = strpbrk(c," \t");
	if (!c1) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLineNo,
			      "%s: %d: Bad line: %.30s..."),
		      filename,lineno,buf);
	    (*errors) ++;
	    continue;
	}
	*c1 = '\0';
	c1++;

	if (!is_ascii_value(filename,lineno,errors,NULL,c))
	    continue; /* Non-ascii hostname */

	while (*c1 && whitespace (*c1)) /* skip leading whitespace */
	    c1++;
	if (!*c1) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLineNo,
			      "%s: %d: Bad line: %.30s..."),
		      filename,lineno,buf);
	    (*errors) ++;
	    continue;
	}
	
	options = strpbrk(c1," \t;");
	if (options) {
	    *options = '\0';
	    options++;
	}


	if (!is_ascii_value(filename,lineno,errors,NULL,c))
	    continue; /* Non-ascii keyword */

	for (st = &SERVICE_TYPES[0]; st->name; st++) 
	    if (0 == istrcmp(c1,st->name))
		break;

	if (!st->name) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeServiceTypeNo,
			      "%s: %d: Bad service type %s"),
		      filename,lineno,c1);
	    (*errors) ++;
	    continue;
	}

	se =  malloc_service_entry(c  /* hostname */,
				 st /* service */,
				 0  /* This is NOT temporary entry */);		
	if (options) {
	    char *opt;
	    char * WALK = NULL;
	    
	    for (opt = mime_parse_content_opts(options, &WALK); 
		 opt; 
		 opt = mime_parse_content_opts(NULL, &WALK)) {
		char * q = strchr(opt,'=');
		char * zopt = NULL;
		char * val = NULL;
		
		DPRINT(Debug,65,(&Debug, 
				 "mime_parse_content_opts gives: %s\n",
				 opt));
		
		if (q) 
		    *q++ = '\0';		    

		if (!is_ascii_value(filename,lineno,errors,NULL,opt))
		    continue; /* Non-ascii keyword */


		if (0 == strcmp(opt,"alias")) {
		    char * conv = NULL;
		    int convlen = 0;
		    if (!q || !q[0]) {
			lib_error(CATGETS(elm_msg_cat, MeSet, MeOptionReqValueNo,
					  "%s: %d: Option %s requires value"),
				  filename,lineno,opt);
			(*errors) ++;
			continue;
		    }

		    if (cs != system_charset) {
			struct string * buffer = convert_se_value(filename,lineno,errors,
								  opt,cs,q,strlen(q));
			if (!buffer)
			    continue;

			convlen =  rc_value_to_syscs(filename,lineno,errors,
						     opt,buffer,&conv);
			if (!conv)
			    continue;

		    } else {
			conv = q;
			convlen = strlen(q);
		    }

		    val = dequote_opt(conv,convlen);		
		    add_alias_to_entry(se,
				       val);
		    free(val);  /* add_alias_to_entry strdup's value */
		    
		    if (conv != q)
			free(conv);
		    se->flags |= SE_given_alias;

		} else if (0 == strcmp(opt,"addr")) {
		    char *q1 = NULL;
		    if (!q || !q[0]) {
			lib_error(CATGETS(elm_msg_cat, MeSet, MeOptionReqValueNo,
					  "%s: %d: Option %s requires value"),
				  filename,lineno,opt);
			(*errors) ++;
			continue;
		    }

		    if (! is_ascii_value(filename,lineno,errors,opt,q))
			continue;

		    if ('"' != q[0])
			q1 = strchr(q,':');

		    if (!q1) { /* name given */
			add_addr_name_to_entry(se,
					       /* Malloces result */
					       dequote_opt(q,strlen(q)),
					       0);
			se->flags |= 
			    SE_given_name_addr;
		    } else {  /* Addr? given */
			SOCKADDR result;
#ifndef USE_INET_ATON
			long tmp;
#endif
			char * port, *tail = "";
			char tail1[2];
			
			tail1[0] = '\0';

			/* Make comparision easier */
			bzero((void *)&result, sizeof result);

			*q1++ = '\0';
			port = qstrpbrk(q1,":/");
			if (port) {
			    tail1[0] = *port;
			    tail1[1] = '\0';
			    *port++ = '\0';
			    tail = port;
			}
			val = dequote_opt(q1,strlen(q1));		

			/* inet_aton() is better but not available on
			   some systems so we need use inet_addr()
			*/

#ifdef I_NETINET_IN			
			if (0 == strcmp(q,"ip") &&
#ifdef USE_INET_ATON
			    inet_aton(val,&result.sin.sin_addr)
#else
			    -1 != (tmp = inet_addr(val)) 
#endif
			    ) {
			    int p = PORT_end;   /* No port given */

#ifndef USE_INET_ATON
			    memcpy(&(result.sin.sin_addr.s_addr),
				   &tmp,
				   sizeof (result.sin.sin_addr.s_addr));
#endif

			    result.sin.sin_family = AF_INET;
			    if (port) {
				char *p2;
				long l;
				
				p = (l = strtol(port,&p2,10));
				if (*p2 != '\0' || p != l) {
				    DPRINT(Debug,10,(&Debug,  "... bad port: %s\n",port));
				    goto addr_bad_port;
				}
			    }
			    result.sin.sin_port = htons(p);
			} else
#endif			
			    {
			    addr_bad_port:

				lib_error(CATGETS(elm_msg_cat, MeSet,
						  MeBadAddrSpecNo,
						  "%s: %d: Bad address specification %s:%s%s%s"),
					  filename,lineno,q,q1,tail1,tail);	
				(*errors) ++;

				DPRINT(Debug,10,(&Debug,  "... dequoted address: %s\n",val));

				free(val); val= NULL;
				continue;			    
			    }	

			add_addr_to_entry(se,&result);
			free(val); val = NULL;
			se->flags |= 
			    SE_given_addr;
		    }
		} else if (0 == strcmp(opt,"port")) {

		    if (!q || !q[0]) {
			lib_error(CATGETS(elm_msg_cat, MeSet, MeOptionReqValueNo,
					  "%s: %d: Option %s requires value"),
				  filename,lineno,opt);
			(*errors) ++;
			continue;
		    }

		    if (! is_ascii_value(filename,lineno,errors,opt,q))
			continue;

		    if (!isdigit(*q)) {

			add_port_name_to_entry(se,
					       /* Malloces result */
					       dequote_opt(q,strlen(q)),
					       0);
			se->flags |= 
			    SE_given_name_port;

		    } else {

			int port;
			char *p2 = NULL;
			long l;


			port = (l = strtol(q,&p2,10));
			if (*p2 != '\0' || port != l) {
			    lib_error(CATGETS(elm_msg_cat, MeSet, MeOptionBad,
					      "%s: %d: Option %s=%s unsupported"),
				      filename,lineno,opt,q);
			    (*errors) ++;
			    continue;
			}
			
			if (port == PORT_end) {
			    lib_error(CATGETS(elm_msg_cat, MeSet, MeOptionPortBadNo,
					      "%s: %d: Option port=%d unsupported"),
				      filename,lineno,port);
			    (*errors) ++;
			    continue;
			}

			add_port_to_entry(se,
					  port);

			se->flags |= SE_given_port;
		    }

		}  else if (0 == strcmp(opt,"verify-tls-certificate")) {

			if (q) {

			    if (! is_ascii_value(filename,lineno,errors,opt,q))
				continue;

			    if (0 == istrcmp(q,"on") || 0 == istrcmp(q,"yes")) {
				if (!(se->flags & 
				      SE_nover_tls_cert))
				    se->flags |= SE_ver_tls_cert;
				else {
				    lib_error(CATGETS(elm_msg_cat, MeSet,MeConfOpt2,
						  "%s: %d: Conflicting option %s=%s"),
					      filename,lineno,opt,q);
				    (*errors) ++;
				}
			    } else  if (0 == istrcmp(q,"no") || 0 == istrcmp(q,"off")) {

				if (!(se->flags & 
				      SE_ver_tls_cert))
				    se->flags 
					|= SE_nover_tls_cert;
				else {
				    lib_error(CATGETS(elm_msg_cat, MeSet,MeConfOpt2,
						  "%s: %d: Conflicting option %s=%s"),
					      filename,lineno,opt,q);
				    (*errors) ++;
				}
			    } else {
				lib_error(CATGETS(elm_msg_cat, MeSet, MeOptionBad,
					  "%s: %d: Option %s=%s unsupported"),
					  filename,lineno,opt,q);
				(*errors) ++;
			    }

			} else {
			    if (!(se->flags & 
				  SE_nover_tls_cert))
				se->flags |= SE_ver_tls_cert;
			    else {
				lib_error(CATGETS(elm_msg_cat, MeSet,MeConfOpt,
						  "%s: %d: Conflicting option %s"),
					  filename,lineno,opt);
				(*errors) ++;
			    }
			}

		}  else if (0 == strcmp(opt,"require-tls-peer-name")) {

		    if (q) {
			char * val0 = dequote_opt(q,strlen(q));
			struct string * val =
			    convert_se_value(filename,lineno,errors,
					     opt,cs,val0,strlen(val0));

			free(val0); val0 = NULL;
			if (!val) 
			    continue;

			add_req_peername_to_entry(se,val);

			free_string(&val);

		    } else
			se->flags |= SE_require_tls_name;

#ifdef USE_DLOPEN
		} else if (NULL != (zopt = strchr(opt,':'))) {
		    int idx = -1;
		    int x;

		    char * conv = NULL;
		    int convlen UNUSED_VAROK = 0;

		    *zopt++ = '\0';

		    for (x = 0; x < se->option_count; x++) {
			if (0 == strcmp(opt,se->
					option_list[x].prefix))
			    idx = x;
		    }
		    if (-1 == idx) {
			struct SE_option_type * Y = get_option_type(opt);
			if (!Y)
			    continue;

			idx = 
			    add_option_type_to_entry(se,
						     Y,opt);
		    }

		    if (SE_option_t_magic != 
			se->option_list[idx].type->magic)
			panic("CONNECTION PANIC",__FILE__,__LINE__,
			      "parse_service_entries",
			      "Bad option type on service list",0);

		    if (cs != system_charset && q) {
			struct string * buffer = convert_se_value(filename,lineno,errors,
								  opt,cs,q,strlen(q));
			if (!buffer)
			    continue;
			
			convlen =  rc_value_to_syscs(filename,lineno,errors,
						     opt,buffer,&conv);
			if (!conv)
			    continue;
		    } else {
			conv = q;
			convlen = strlen(q);
		    }

		    if (!se->option_list[idx].type ->
			parse_on_option(& (se->
					   option_list[idx]),
					zopt,conv)) {
			(*errors) ++;

			if (conv && conv != q)
			    free(conv);

			continue;
		    }					   
		    
		    if (conv && conv != q)
			free(conv);
		    
#endif
		} else {
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeOptionUnspportedNo,
				      "%s: %d: Option %s unsupported"),
			      filename,lineno,opt);
		    (*errors) ++;
		    continue;
		}
	    }

	}

	if (0 != (st->flags & STFLAG_require_port) &&
	    0 == (se->flags & SE_given_name_port) &&
	    0 == (se->flags & SE_given_port)) {
	    
	    int have_port = 0;

	    if (0 != (se->flags & SE_given_addr)) {	       
		have_port = addrs_have_port(se);
	    }

	    if (!have_port) {
		lib_error(CATGETS(elm_msg_cat, MeSet, 
				  MeServiceTypeReqPort,
				  "%s: %d: Service type %s requires port= or addr= with port definition"),
			  filename,lineno, st->name);
		(*errors) ++;
		goto free_it;
	    }
	}
	
	ret->service_list[ret->service_count] = se;
	inc_service_entry_refcount(ret->service_list[ret->service_count++]);

    free_it:
	free_service_entry(&se);
    }

    if (fileset)
	*fileset = cs;

    DPRINT(Debug,9,(&Debug, 		    
		    "parse_service_entries: %s, %d parsed\n",
		    filename,ret->service_count));
    fclose(f);

    if (buf)
	free(buf);

    return ret;
}

#endif


/* Returns domain name if matches to special-use-domains-blacklist */
const char * is_special_use_domain(host_domain_name)
     const char * host_domain_name;
{
     /* Pointer to internal list, do NOT free */
    const char ** blacklist = 
	give_dt_path_as_elems(&special_bl_domains,
			      "special-use-domains-blacklist");

    /* If domain name is given as ".domain" then it blacklists names under 
       that domain. If name is given as "domain" then also "domain" itself 
       (and names under it) are blacklisted. 
    */

    const char * p;
    int i;

    if (!blacklist)
	return NULL;

    /* These domains fail on normal processing */
    for (i = 0; blacklist[i]; i++)
	if (0 == istrcmp(host_domain_name,blacklist[i]))
	    return blacklist[i];

    /* look names under these domains */

    for (p = host_domain_name; *p; p++) {
	if ('.' == *p) {
	    for (i = 0; blacklist[i]; i++) {
		if (0 == istrcmp(p+1,blacklist[i]))
		    return blacklist[i];
		if (0 == istrcmp(p,blacklist[i]))
		    return blacklist[i];
	    }
	}
    }
    
    return NULL;    
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */

